/***********************************************************************
 * Copyright (c) 2003 University of Kent.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.cs.kent.ac.uk/projects/ocl/CPL.html
 ***********************************************************************/

package uk.ac.kent.cs.ocl20.standard.lib;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

public class OclTupleImpl extends OclAnyImpl implements OclTuple {
	private java.util.Map _implementation;

	protected java.util.Map implementation() {
		return _implementation;
	}

	protected OclTupleImpl(Map m, StdLibAdapter adapter) {
		super(adapter);
		_implementation = m;
	}

	protected OclTupleImpl(OclString[] keys, OclAny[] arr, StdLibAdapter adapter) {
		super(adapter);
		_implementation = new HashMap();
		for (int i = 0; i < keys.length; i++) {
			_implementation.put(keys[i], arr[i]);
		}
	}
	public Object asJavaObject() {
		return implementation();
	}

	//--- IOclTuple ---
	public OclAny property(OclString s) {
		OclString key = s;
		OclAny o = (OclAny)_implementation.get(key);
		if (o == null)
			//return adapter.Undefined("Tuple doesn't contain element - "+key,null);
			return adapter.Undefined();
		return o;
	}

	public OclAny property(OclString s, Object[] pobjs) {
		return adapter.Undefined(); //property(s);
	}

	public void setProperty(OclString name, Object value) {
		_implementation.put(name, value);
	}

	//--- Object ---
	public String toString() {
		String str = "Tuple { ";
		Iterator i = _implementation.keySet().iterator();
		while (i.hasNext()) {
			OclString s = (OclString) i.next();
			str += adapter.impl(s) + " = " + _implementation.get(s);
			if (i.hasNext())
				str += ", ";
		}
		return str + " }";
	}

	public boolean equals(Object o) {
		return (o instanceof OclTuple) ? adapter.impl(equalTo((OclTuple) o)).booleanValue() : false;
	}

	public Object clone() {
		return new OclTupleImpl(
			(OclString[])_implementation.keySet().toArray(new OclString[0]),
			(OclAny[])_implementation.values().toArray(new OclAny[0]),
			super.adapter
			);
	}

	//--- IOclVisitable ---
	public Object accept(OclVisitor v, Object obj) {
		return v.visit(this, obj);
	}
}
