/***********************************************************************
 * Copyright (c) 2003 University of Kent.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.cs.kent.ac.uk/projects/ocl/CPL.html
 ***********************************************************************/

/**
 * 
 * @author Octavian Patrascoiu
 *
 */

package uk.ac.kent.cs.ocl20.synthesis;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import uk.ac.kent.cs.kmf.util.ILog;
import uk.ac.kent.cs.kmf.util.OutputStreamLog;
import uk.ac.kent.cs.ocl20.OclProcessor;
import uk.ac.kent.cs.ocl20.semantics.model.contexts.ContextDeclaration;

public class OclEvaluatorImpl
implements OclEvaluator
{
	/** Constructor */
	public OclEvaluatorImpl(OclProcessor processor, OclEvaluatorVisitorImpl evaluator) {
		this.processor = processor;
		this.evaluator = evaluator;
	}

	OclProcessor processor;
	OclEvaluatorVisitorImpl evaluator;
	
	/** Input semantic tree */
	protected ContextDeclaration decl;
	
	/** Runtime environment */
	protected RuntimeEnvironment renv;
	
	/** Log file */
	protected ILog log;
	
	/** Evaluate contexts into the semantic and runtime environment */
	protected List evaluate() {
		//--- Check evaluation context ---
		if (decl == null) return null;
		if (renv == null) {
			renv = new RuntimeEnvironmentImpl();
		}
		if (log == null) log = new OutputStreamLog(System.out);
		
		//--- Create the context for the evalution --- 
		Map data = new HashMap();
		data.put("env", renv);
		data.put("log", log);

		//--- Evaluate the input ---
		List result = (List)decl.accept(evaluator, data);
		
		//--- Return result ---
		return result;
	}
	
	/** Evaluate contexts into the semantic and runtime environment */
	public List evaluate(ContextDeclaration decl, RuntimeEnvironment renv) {
		//--- Set local variables ---
		this.decl = decl;
		this.renv = renv;
		this.log = new OutputStreamLog(System.out);
		
		//--- Evaluate ---
		return evaluate();
	}

	/** Evaluate contexts into the semantic and runtime environment */
	public List evaluate(ContextDeclaration decl) {
		//--- Set local variables ---
		this.decl = decl;
		this.renv = new RuntimeEnvironmentImpl();
		renv.setValue("self", null);
		this.log = new OutputStreamLog(System.out);
		
		//--- Evaluate ---
		return evaluate();
	}

	/** Evaluate contexts into the semantic and runtime environment */
	public List evaluate(ContextDeclaration decl, RuntimeEnvironment renv, ILog log) {
		//--- Set local variables ---
		this.decl = decl;
		this.renv = renv;
		this.log = log;
		
		//--- Evaluate ---
		return evaluate();
	}
}
